package de.tub.mapc.Coordination;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.slf4j.Logger;

import de.dailab.keks.io.IAddress;
import de.jiac.micro.core.IAgent;
import de.jiac.micro.core.scope.AgentScope;
import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.Coordination.goals.AbstractGoal;
import de.tub.mapc.Coordination.goals.GoalExploreField;
import de.tub.mapc.Coordination.goals.GoalFindSwitchForFence;
import de.tub.mapc.Coordination.goals.GoalGoToSwitchingPosition;
import de.tub.mapc.Coordination.messages.SwitchingAnswer;
import de.tub.mapc.knowledge.Gridworld;
import de.tub.mapc.knowledge.Perception;
import de.tub.mapc.knowledge.Position;
import de.tub.mapc.knowledge.Switch;
import de.tub.mapc.lang.IntHashMap;
import de.tub.mapc.worldmodel.GridField;



public class CoordinationData {
	
	/**
	 * the number of the current step
	 */
	private int stepNo;
	/**
	 * 
	 * the small worldmodel every agent has
	 */
	private Gridworld world;
		
	
	/**
	 * all known agents, the key is the String representation of the address
	 */
	private HashMap<String, Agent> agents;
	/**
	 * all known Switches, the key is the id of the switch (defined by Position)
	 */
	private IntHashMap<Switch> switches;
	
	/**
	 * this are NOT switches, but fenceFields, where the switch is unknown<br/>
	 * there can be votes for these fenceFields to be explored
	 */
	private List<Switch> unknownSwitches;
	/**
	 * contains switches where an agent is assigned from an switch request
	 * this agent has to get the goal in the next goalassignment
	 */
	private List<SwitchingAnswer> switchesToBeAssignedFromRequest;
	
	/**
	 * the generated points to explore, the key is the id of the Position
	 */
	private IntHashMap<ExploringPoint> pointsToExplore;
	/**
	 * count of perceptions the coordinator received in this step
	 */
	private int receivedPerceptions;

	private Position enemyCorralUpperLeft;
	private Position enemyCorralLowerRight;
	private boolean foundEnemyCorral;
	
	/**
	 * reference to the agent having the coordination component
	 */
	
	private IAgent _agent;
	private Logger _logger;
	
	public CoordinationData(Gridworld world){
		setStepNo(0);
		agents = new HashMap<String, Agent>();
		switches = new IntHashMap<Switch>();
		unknownSwitches = new ArrayList<Switch>();
		switchesToBeAssignedFromRequest = new ArrayList<SwitchingAnswer>();
		_agent = AgentScope.getAgentReference();
		_logger = _agent.getLogger("CoordinationData");
		this.world = world;
		generatePointsToExplore();
		enemyCorralLowerRight = new Position(0, 0);
		enemyCorralUpperLeft = new Position(9999, 9999);
		foundEnemyCorral = false;
	}
	
	public void reset(){
		//TODO prepare for new Simulation - DONE
		setStepNo(0);
		switches.clear();
		agents.clear();
		unknownSwitches.clear();
		switchesToBeAssignedFromRequest.clear();
		this.world = null;
	}
	
	private void generatePointsToExplore(){
		pointsToExplore = new IntHashMap<ExploringPoint>();
//		int minabstand = world.getLineOfSight()-1;
		int minabstand = 3;
		for(int x = 0; x < world.getGsizeX(); x += minabstand){
			for(int y = 0; y < world.getGsizeY(); y += minabstand){
				ExploringPoint p = new ExploringPoint(x, y);
				pointsToExplore.put(p.getId(), p);
			}
		}
	}
		
	public void addAgent(Agent a){
		agents.put(a.getAddress().toString(), a);
	}
	
	public Agent getAgent(IAddress address){
		return agents.get(address.toString());
	}
	
	public Agent getAgent(String address){
		return agents.get(address);
	}
	
	public int getReceivedPerceptions(){
		return receivedPerceptions;
	}
	
	public HashMap<String, Agent> getAgents(){
		return agents;
	}
	
	
	public void resetReceivedPerceptions(){
		receivedPerceptions = 0;
	}

	public IntHashMap<ExploringPoint> getPointsToExplore() {
		return pointsToExplore;
	}
	
	public List<Switch> getUnknownSwitches() {
		return unknownSwitches;
	}
	
	public List<SwitchingAnswer> getSwitchesToBeAssignedFromRequest() {
		return switchesToBeAssignedFromRequest;
	}
	
	public void addSwitchToBeAssignedFromRequest(SwitchingAnswer s){
		switchesToBeAssignedFromRequest.add(s);
		// TODO wie und wann werden die entfernt? Nach zuteilung des goals
	}
	
	public void addUnknownSwitch(Switch s){
		unknownSwitches.add(s);
		// TODO wie und wann werden die entfernt? wenn der erkunder reachedGoal sendet?
		// wenn die nicht entfernt werden ist zwar nicht schlimm, weil keiner mehr dafuer
		// voted, aber sie sind eben unnuetz, verbrauchen speicher und zeit
	}

	public Gridworld getWorld() {
		return world;
	}
	
	
	public IntHashMap<Switch> getSwitches() {
		return switches;
	}

	public void update(Perception p, IAddress sender){		
		//sometimes updatemessage arrives before initmessage
		if (world == null) {
			_logger.error("CoordinationData update() error: world == null");
			return;  
		}
		// update world
		world.update(p.getGridInfo());
		// update switchlist
		GridField[] fields = p.getGridInfo();
		// Cows found in this perception
		ArrayList<GridField> cows = new ArrayList<GridField>();		
		for(GridField f : fields){
			
			if(pointsToExplore.get(f.getId()) != null){
				// remove this point from exploring list 
				pointsToExplore.remove(f.getId());
			}
			if (f.getType() == IMAPCConstants.TYPE_SWITCH){
				// found switch, check, wheter it is already in the list
				if(switches.get(f.getId()) == null){
					// not in the list, insert
					Switch s = new Switch(f.getX(), f.getY());
					switches.put(s.getId(), s);
				}
			} else if (f.getType() == IMAPCConstants.TYPE_COW){
				cows.add(f);
			}else if (f.getType() == IMAPCConstants.TYPE_ENEMY_CORRAL){
				foundEnemyCorral = true;
				if(f.getX() < enemyCorralUpperLeft.getX()){
					enemyCorralUpperLeft.setX(f.getX());
				}
				if(f.getX() > enemyCorralLowerRight.getX()){
					enemyCorralLowerRight.setX(f.getX());
				}
				if(f.getY() < enemyCorralUpperLeft.getY()){
					enemyCorralUpperLeft.setY(f.getY());
				}
				if(f.getY() > enemyCorralLowerRight.getY()){
					enemyCorralLowerRight.setY(f.getY());
				}
			}
			
		}
		
		// update agents
		Agent a = agents.get(sender.toString());
		if (a == null){
			// put the agent into the hashmap
			agents.put(sender.toString(), new Agent(p.getPosX(),p.getPosY(),sender));
		}else{
			a.setPosition(p.getPosition());
			
		}

		receivedPerceptions++;
	}
	
		
	/**
	 * return a list with nearby(3 fields) cows
	 * @param cow
	 * @return a list with nearby cows
	 */
	/*private ArrayList<GridField> getNearbyCows(GridField cow){
		ArrayList<GridField> nearCows = new ArrayList<GridField>();
		// Search for cows in maximal 3 fields away from this one
		int NEARDISTANCE = 3;
		
		int minX = cow.getX()-NEARDISTANCE;
		int maxX = cow.getX()+NEARDISTANCE;
		int minY = cow.getY()-NEARDISTANCE;
		int maxY = cow.getY()+NEARDISTANCE;
		
		// Test boundaries, limits can be out of the map
		if (minX < 0) minX = 0;
		if (minY < 0) minY = 0;
		if (maxX > world.getGsizeX()) maxX = world.getGsizeX()-2;
		if (maxY > world.getGsizeY()) maxY = world.getGsizeY()-2;
		
		GridField[][] grid = world.getGrid();
		for (int i = minX; i <= maxX; i++){
			for (int j = minY; j <= maxY; j++){
				if ((i==cow.getX()) && (j==cow.getY())){
					// ignore itself
					continue;
				}				
				if (grid[i][j].getType() == IMAPCConstants.TYPE_COW){
					nearCows.add(grid[i][j]);
				}
			}
		}
		
		return nearCows;
	}
	*/
	@SuppressWarnings("deprecation")
	public void setNewGoal(Agent a, AbstractGoal g){
		//TODO andere goaltypen ergaenzen
		
		// ERKLAERUNG: 
		// ein agent war explorer
		// -> er war bei dem exploring point eingetragen
		// dann wird er switcher
		// dann wird der exploring point von jmd anders entdeckt
		// er steht aber noch als zustaendig drin
		// also bekommt er nen neues goal
		// da er aber auch noch im switch steht, geht auch niemand anders zu dem switch
		// ich hab ewig gebraucht um drauf zu kommen ;)
		
		AbstractGoal oldGoal = a.getGoal(); 
		
		if(oldGoal != null && oldGoal instanceof GoalExploreField){
			GoalExploreField og = (GoalExploreField)oldGoal;
			if(getPointsToExplore().get(og.getGoalPosition().getId()) != null){
				getPointsToExplore().get(og.getGoalPosition().getId()).setAgent(null);
			}
		}else if(oldGoal != null && oldGoal instanceof GoalGoToSwitchingPosition){
			GoalGoToSwitchingPosition og = (GoalGoToSwitchingPosition)oldGoal;
			if(getSwitches().get(og.getSwitchPosition().getId()) != null ){
				_logger.trace("removed agent from switch "+og.getSwitchPosition());
				getSwitches().get(og.getSwitchPosition().getId()).setSwitchingAgent("");
			}
		}else if(oldGoal != null && oldGoal instanceof GoalFindSwitchForFence){
			GoalFindSwitchForFence og = (GoalFindSwitchForFence)oldGoal;
			Switch unknownSwitch = getUnknwonSwitchForPosition(og.getFenceField());
			if(unknownSwitch != null){
				_logger.trace("removed agent from switch exploring "+unknownSwitch);
				unknownSwitch.setSwitchingAgent("");
			}
		}
		a.setGoal(g);
	}
	
	private Switch getUnknwonSwitchForPosition(Position p){
		for(Switch current : unknownSwitches){
			if(p.getId() == current.getId()){
				return current;
			}	
		}
		return null;
	}

	public void setStepNo(int stepNo) {
		this.stepNo = stepNo;
	}

	public int getStepNo() {
		return stepNo;
	}

	public Position getEnemyCorralUpperLeft() {
		return enemyCorralUpperLeft;
	}

	public Position getEnemyCorralLowerRight() {
		return enemyCorralLowerRight;
	}

	public boolean isFoundEnemyCorral() {
		return foundEnemyCorral;
	}

}
