package de.tub.mapc.components.contestagents;

import java.awt.Point;
import java.util.LinkedList;
import java.util.List;
import java.util.PriorityQueue;

import org.slf4j.Logger;

import de.jiac.micro.core.IAgent;
import de.jiac.micro.core.scope.AgentScope;
import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.components.contestagents.handle.IPathFindingHandle;
import de.tub.mapc.knowledge.Gridworld;
import de.tub.mapc.knowledge.Position;
import de.tub.mapc.lang.IntHashMap;
import de.tub.mapc.lang.IntHashSet;
import de.tub.mapc.monitor.AgentMonitor;
import de.tub.mapc.pathfinding.AStarNode;
import de.tub.mapc.pathfinding.PathMap;
import de.tub.mapc.pathfinding.PathResult;
import de.tub.mapc.worldmodel.GridField;

/*
 *   Changed to an A* algorithm, since we don't need a D* (25.06.2010)
 */

public class PathFindingElement implements IPathFindingHandle {
//	/**
//	 * Alpha can be between 0 (BFS) and 1 (Dijkstra's), with 0.5 being A*
//	 */
//	private static double ALPHA = 0.4999D;
//
//	private static double ONE_MINUS_ALPHA = 1D - ALPHA;
//
//	private static double ALPHA_NORM = Math.max(ALPHA, ONE_MINUS_ALPHA);

	/**
	 * All nodes of the grid
	 */
	private IntHashMap<AStarNode> _allNodes;

	private AStarNode Astart, Agoal;
	/**
	 * The world grid
	 */
	private GridField[][] grid;

	private boolean isInitialized = false;
	private boolean ignoreFences = true;
	private boolean ignoreCows = false;
	private boolean circleHerds = false;
	private int herdleftboundary;
	private int herdrightboundary;
	private int herdupperboundary;
	private int herdlowerboundary;

	// Change to true to show monitor with this agent's path
	private boolean showMonitor = true;

	private Logger _logger;

	// MONITOR
	private PathMap pathMap;
	private int tabNumber;
	private boolean monitorUp = false;

	public PathFindingElement() {

	}

	/**
	 * creates a new Pathfinder
	 * 
	 * @param show
	 *            if true, a new tab in Monitor is created for this path
	 */
	public PathFindingElement(boolean show) {
		showMonitor = show;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public PathResult getPath(Position goal, Position start, Gridworld world,
			boolean ignoreFences) {
		this.ignoreFences = ignoreFences;
		this.ignoreCows = false;
		this.circleHerds = false;
		return getAStarPath(goal.getX(), goal.getY(), start.getX(),
				start.getY(), world);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public PathResult getPathForHerd(Position goal, Position start,
			Gridworld world) {
		this.ignoreFences = true;
		this.ignoreCows = true;
		this.circleHerds = false;
		return getAStarPath(goal.getX(), goal.getY(), start.getX(),
				start.getY(), world);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public PathResult getPathForBlocker(Position goal, Position start,
			Gridworld world) {
		this.ignoreFences = true;
		this.ignoreCows = true;
		this.circleHerds = false;
		return getAStarPath(goal.getX(), goal.getY(), start.getX(),
				start.getY(), world);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public PathResult getPathForDriver(Position goal, Position start,
			Gridworld world, Point ul, Point lr) {
		this.ignoreFences = true;
		this.ignoreCows = false;
		this.circleHerds = true;
		herdleftboundary = ul.x;
		herdrightboundary = lr.x;
		herdupperboundary = ul.y;
		herdlowerboundary = lr.y;
		return getAStarPath(goal.getX(), goal.getY(), start.getX(),
				start.getY(), world);
	}

	private void reset() {
		for (AStarNode node : _allNodes.valuesAsArray(AStarNode.class)) {
			node.reset();
		}
	}

	private void showPathOnMonitor(List<AStarNode> path) {
		if (showMonitor) {
			if (!monitorUp) {
				int r = initWindow();
				if (r < 0) {
					_logger.warn("PATH_MONITOR: Monitor was null, not added");
				}
			}
			pathMap.update(grid, path, Astart, Agoal);
			// pathMap.update(grid, path);
		}
	}

	private PathResult getAStarPath(int goalx, int goaly, int x, int y,
			Gridworld world) {
		// Find path with AStar
		
		PathResult result = new PathResult();
		if (goalx < 0 || goaly < 0) {
			_logger.warn("PATH: ERROR IN PATH FINDING! goal with negative ");
			result.setErrortype(IMAPCConstants.PATH_ERROR);
			return result;
		}
		if (!isInitialized && !init(world)) {
			result.setErrortype(IMAPCConstants.PATH_ERROR);
			return result;
		} else {
			// Calculated a new path in this step
			// Reset everything needed
			reset();

			Astart = findNode(x, y);
			if (Astart == null) {
				_logger.error("PATH: Unknown start position: (" + x + "," + y
						+ ")");
				// for (int i=0; i<
				// Thread.currentThread().getStackTrace().length; i++)
				// System.out.println(Thread.currentThread().getStackTrace()[i].getClassName()+" "+Thread.currentThread().getStackTrace()[i].getMethodName()+
				// " "+Thread.currentThread().getStackTrace()[i].getLineNumber());
				result.setErrortype(IMAPCConstants.PATH_ERROR);
				return result;
			}

			Agoal = findNode(goalx, goaly);
			if (Agoal == null) {
				_logger.error("PATH: Goal not found: (" + goalx + "," + goaly
						+ ")");
				result.setErrortype(IMAPCConstants.PATH_ERROR);
				return result;
			}

			if (Astart.equals(Agoal)) {
				// We are already at the goal
				// showPathOnMonitor(path);
				result.setReachedGoal(true);
				return result;
			}

			// Calculate path
			// _logger.trace("PATH: Starting Calculations");
			LinkedList<AStarNode> path = AStarAlgorithm(Astart, Agoal);
			if (path == null) {
				result.setErrortype(IMAPCConstants.PATH_NOTFOUND);
				return result;
			}
			// _logger.trace("PATH: Finished Calculations");
			// _logger.info("PATH: My goal is (" + g.getPosx() + "," +
			// g.getPosy() + ")");

			showPathOnMonitor(path);

			// save length of the path
			result.setPathLength(path.size());

			int stepsToNextFence = 1;
			boolean fenceFound = false;
			for (AStarNode next : path) {
				if ((grid[next.getX()][next.getY()].getType() == IMAPCConstants.TYPE_FENCE)
						|| (grid[next.getX()][next.getY()].getType() == IMAPCConstants.TYPE_OPEN_FENCE)) {
					Point p = new Point(next.getX(), next.getY());
					result.addFence(p);

					if (!fenceFound) {
						// This is the first fence
						result.setStepsToNextFence(stepsToNextFence);
						result.setFencex(p.x);
						result.setFencey(p.y);
						result.setErrortype(IMAPCConstants.PATH_FENCEAHEAD);
					}

					// _logger.debug("PATH: There is a fence in: "+p.x+" "+p.y);
					fenceFound = true;
				}
				stepsToNextFence++;
			}

			if (path.size() < 1) {
				_logger.warn("PATH: ERROR IN PATH FINDING! 1");
				result.setErrortype(IMAPCConstants.PATH_ERROR);
				return result;
			}

			// The first should be the best next move
			AStarNode next = path.getFirst();

			GridField gf = grid[next.getX()][next.getY()];
			if (isBlocked(gf)) {
				result.setErrortype(IMAPCConstants.PATH_BLOCKED);
				return result;
			}
			if (gf.getType() == IMAPCConstants.TYPE_FENCE) {
				result.setErrortype(IMAPCConstants.PATH_FENCEAHEAD);
				result.setFencex(gf.getX());
				result.setFencey(gf.getY());
				result.setThisStep(IMAPCConstants.ACTION_SKIP);
				return result;
			}

			// get next Positions for computing driving positions
			if (path.size() > 2) {
				AStarNode n = path.get(2);
				result.setThreeStepsPosition(n);
			}

			// positions relative to the agent position
			int posx = (next.getX() - Astart.getX());
			int posy = (next.getY() - Astart.getY());
			// _logger.trace("PATH: Relative pos: (" + posx + "," + posy + ")");
			if (posx == 0) {
				if (posy == -1)
					result.setThisStep(IMAPCConstants.ACTION_NORTH);
				if (posy == 1)
					result.setThisStep(IMAPCConstants.ACTION_SOUTH);
				return result;
			}
			if (posx == 1) {
				if (posy == -1)
					result.setThisStep(IMAPCConstants.ACTION_NORTHEAST);
				if (posy == 0)
					result.setThisStep(IMAPCConstants.ACTION_EAST);
				if (posy == 1)
					result.setThisStep(IMAPCConstants.ACTION_SOUTHEAST);
				return result;
			}
			if (posx == -1) {
				if (posy == -1)
					result.setThisStep(IMAPCConstants.ACTION_NORTHWEST);
				if (posy == 0)
					result.setThisStep(IMAPCConstants.ACTION_WEST);
				if (posy == 1)
					result.setThisStep(IMAPCConstants.ACTION_SOUTHWEST);
				return result;
			}

			_logger.warn("returning ACTION_SKIP");
			return new PathResult(IMAPCConstants.ACTION_SKIP);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int restart() {
		// next time we calculate a path, init() will be called in the beginning
		isInitialized = false;

		if (showMonitor)
			monitorUp = false;

		return 0;
	}

	/**
	 * returns true if the field is blocked
	 * 
	 * @param gridfield
	 * @return
	 */
	private boolean isBlocked(GridField gridfield) {
		final int type = gridfield.getType();

		if ((type == IMAPCConstants.TYPE_TREE)
				|| (type == IMAPCConstants.TYPE_ENEMY)
				|| (type == IMAPCConstants.TYPE_PARTNER)
				|| (type == IMAPCConstants.TYPE_SWITCH)) {
			return true;
		}

		if (type == IMAPCConstants.TYPE_FENCE && !ignoreFences) {
			return true;
		}

		if (type == IMAPCConstants.TYPE_COW && !ignoreCows) {
			return true;
		}

		return false;
	}

	// For AStar
	private AStarNode findNode(int x, int y) {
		// manually inlined @see Position.getIDForXY()
		return _allNodes.get((x << 16) | y);
	}

	private boolean init(Gridworld world) {
		// Referenz auf den Agenten
		IAgent agent = AgentScope.getAgentReference();
		// Referenz auf den Logger
		_logger = agent.getLogger("PathFinder");

		// avoid resizing of map
		final int maxNodes = world.getGsizeX() * world.getGsizeY();
		_allNodes = new IntHashMap<AStarNode>(maxNodes);

		grid = world.getGrid();

		// For the AStar
		// Build the nodes
		for (int i = 0; i < world.getGsizeX(); i++) {
			for (int j = 0; j < world.getGsizeY(); j++) {
				AStarNode n = new AStarNode(i, j);
				_allNodes.put(n.getId(), n);
			}
		}

		isInitialized = true;
		return true;

	}

	// MONITOR
	private int initWindow() {
		AgentMonitor inst = AgentMonitor.getInstance();
		if (inst == null)
			return -1;

		pathMap = new PathMap();
		tabNumber = inst.addTab(pathMap, AgentScope.getAgentReference()
				.toString());
		if (tabNumber >= 0)
			monitorUp = true;

		return tabNumber;
	}

	/**
	 * Return the path until node (without the starting position)
	 * 
	 * @param node
	 * @return
	 */
	private LinkedList<AStarNode> reconstructPath(AStarNode firstNode) {
		LinkedList<AStarNode> path = new LinkedList<AStarNode>();
		AStarNode node = firstNode;
		while (node.pathParent != null) {
			path.addFirst(node);
			node = node.pathParent;
		}
		return path;
	}

	private double heuristic(AStarNode node, AStarNode goal) {
		double h = Math.max(Math.abs(node.getX() - goal.getX()),
				Math.abs(node.getY() - goal.getY()));

		double dx1 = node.getX() - goal.getX();
		double dy1 = node.getY() - goal.getY();
		double dx2 = Astart.getX() - goal.getX();
		double dy2 = Astart.getY() - goal.getY();
		double cross = Math.abs(dx1 * dy2 - dx2 * dy1);
		h += cross * 0.001;

		return h;
	}

	private int getNeighbors(AStarNode node, AStarNode[] neighbors) {
		final int x = node.getX();
		final int y = node.getY();

		// Below are the ids of the neighbors
		// x+1+"#"+y;
		// x+1+"#"+y-1;
		// x+1+"#"+y+1;
		// x-1+"#"+y;
		// x-1+"#"+y-1;
		// x-1+"#"+y+1;
		// x+"#"+y-1;
		// x+"#"+y+1;

		int index = 0;
		for (int i = -1; i <= 1; i++) {
			for (int k = -1; k <= 1; k++) {
				if (i == 0 && k == 0) {
					continue; // jump the State s, get only the neighbors
				}

				AStarNode aux = findNode(x + i, y + k);
				if (aux == null) {
					continue; // state not found (position is probably out of
								// bounds)
				}

				if (!isBlocked(grid[aux.getX()][aux.getY()])) {
					neighbors[index++] = aux;
				}
			}
		}

		return index;
	}

	private double distance(AStarNode from, AStarNode to) {
		final int toX = to.getX();
		final int toY = to.getY();

		GridField gridfield = grid[toX][toY];

		if (circleHerds) {
			// the fields in and around the herd have higher cost, so the agents
			// won't go through herds
			if (toX >= (herdleftboundary - 3) && toX <= (herdrightboundary + 3)
					&& toY >= (herdupperboundary - 3)
					&& toY <= (herdlowerboundary + 3)) {
				return 10;
			}
		}
		
		// fix for the switch problem
		if(gridfield.getType() == IMAPCConstants.TYPE_FENCE) {
		    final int diffx= from.getX() - toX;
		    final int diffy= from.getY() - toY;
		    
		    if(diffx != 0 && diffy != 0) {
		        // diagonally move -> check for switch in between
		        
		        if(grid[toX + diffx][toY].getType() == IMAPCConstants.TYPE_SWITCH || grid[toX][toY + diffy].getType() == IMAPCConstants.TYPE_SWITCH) {
		            // agent should do a straight move
		            return 2;
		        }
		    }
		}

		// agents should better not walk through the own corral
		if (gridfield.getType() == IMAPCConstants.TYPE_CORRAL)
			return 10;

		// _logger.debug("Empty on: "+dest.getPosx()+","+dest.getPosy());
		return 1.0;
	}

	final PriorityQueue<AStarNode> openNodes = new PriorityQueue<AStarNode>();
	final IntHashSet openKeys = new IntHashSet();
	final IntHashMap<AStarNode> closedNodes = new IntHashMap<AStarNode>();
	final AStarNode[] neighbors = new AStarNode[8];

	/**
	 * Return the path from start to goal, or null if there is not path
	 * 
	 * @param start
	 *            - the starting position
	 * @param goal
	 *            - the goal position
	 * @return the path as a list of nodes, or null if there is not path
	 */
	private LinkedList<AStarNode> AStarAlgorithm(AStarNode start, AStarNode goal) {
//		long time = System.nanoTime();
		
		start.costFromStart = 0;
		start.estimatedCostToGoal = heuristic(start, goal);
		start.pathParent = null;

		// avoid field accesses, it's faster that way
		openNodes.clear();
		openKeys.clear();
		closedNodes.clear();

		openNodes.add(start);
		openKeys.add(start.getId());

		while (!openNodes.isEmpty()) {
			// Get the node with the lowest F value
			AStarNode currentNode = openNodes.remove();
			openKeys.remove(currentNode.getId());

			// If we are at the goal then return the path to it
			if (currentNode.getId() == goal.getId()) {
				LinkedList<AStarNode> path = reconstructPath(currentNode);
//				time = System.nanoTime() - time;
//				System.out.println("Path found! computation time: " + time + " ns");
				return path;
			}

			// Add the neighbors of this node to open
			int neighborCount = getNeighbors(currentNode, neighbors);
			for (int n = 0; n < neighborCount; ++n) {
				final AStarNode nearNode = neighbors[n];

				double costFromStart = currentNode.costFromStart
						+ distance(currentNode, nearNode);
				boolean isClosed = closedNodes.containsKey(nearNode.getId());
				boolean isOpen = openKeys.contains(nearNode.getId());

				if ((!isClosed && !isOpen)
						|| costFromStart < nearNode.costFromStart) {
					nearNode.pathParent = currentNode;
					nearNode.costFromStart = costFromStart;
					nearNode.estimatedCostToGoal = heuristic(nearNode, goal);
					if (isClosed) {
						closedNodes.remove(nearNode.getId());
					}
					if (!isOpen) {
						openKeys.add(nearNode.getId());
						openNodes.add(nearNode);
					}
				}
			}

			closedNodes.put(currentNode.getId(), currentNode);
		}

		// PATH NOT FOUND
//		time = System.nanoTime() - time;
//		System.out.println("Path not found! computation time: " + time + " ns");
		return null;
	}
}
